#!/usr/bin/env python
#
# meshopts.py - The MeshOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`MeshOpts` class, which defines settings
for displaying :class:`.TriangleMesh` overlays.
"""


import logging

import numpy as np

import props

import fsl.data.image       as fslimage
import fsl.utils.transform  as transform

import fsleyes.colourmaps   as colourmaps
import fsleyes.overlay      as fsloverlay
import fsleyes.colourmaps   as fslcmaps
from . import display       as fsldisplay
from . import colourmapopts as cmapopts


log = logging.getLogger(__name__)



def genMeshColour(overlay):
    """Called by :meth:`MeshOpts.__init__`. Generates an initial colour for
    the given :class:`.TriangleMesh` overlay.

    If the overlay file name looks like it was generated by the FSL FIRST
    segmentation tool, returns a colour from the ``mgh-cma-freesurfer`` colour
    map. Otherwise returns a random colour.
    """
    filename        = str(overlay.dataSource)
    subcorticalCmap = colourmaps.getLookupTable('mgh-cma-freesurfer')

    if   'L_Thal' in filename: return subcorticalCmap.get(10).colour
    elif 'L_Caud' in filename: return subcorticalCmap.get(11).colour
    elif 'L_Puta' in filename: return subcorticalCmap.get(12).colour
    elif 'L_Pall' in filename: return subcorticalCmap.get(13).colour
    elif 'BrStem' in filename: return subcorticalCmap.get(16).colour
    elif 'L_Hipp' in filename: return subcorticalCmap.get(17).colour
    elif 'L_Amyg' in filename: return subcorticalCmap.get(18).colour
    elif 'L_Accu' in filename: return subcorticalCmap.get(26).colour
    elif 'R_Thal' in filename: return subcorticalCmap.get(49).colour
    elif 'R_Caud' in filename: return subcorticalCmap.get(50).colour
    elif 'R_Puta' in filename: return subcorticalCmap.get(51).colour
    elif 'R_Pall' in filename: return subcorticalCmap.get(52).colour
    elif 'R_Hipp' in filename: return subcorticalCmap.get(53).colour
    elif 'R_Amyg' in filename: return subcorticalCmap.get(54).colour
    elif 'R_Accu' in filename: return subcorticalCmap.get(58).colour
    
    return colourmaps.randomBrightColour()


class MeshOpts(cmapopts.ColourMapOpts, fsldisplay.DisplayOpts):
    """The ``MeshOpts`` class defines settings for displaying
    :class:`.TriangleMesh` overlays. See also the :class:`.GiftiOpts`
    sub-class.
    """


    colour = props.Colour()
    """The mesh colour. """

    
    outline = props.Boolean(default=False)
    """If ``True``, an outline of the mesh is shown. Otherwise a
    cross- section of the mesh is filled.
    """

    
    outlineWidth = props.Real(minval=0.1, maxval=10, default=2, clamped=False)
    """If :attr:`outline` is ``True``, this property defines the width of the
    outline in pixels.
    """


    showName = props.Boolean(default=False)
    """If ``True``, the mesh name is shown alongside it.

    .. note:: Not implemented yet, and maybe never will be.
    """


    vertexData = props.Choice((None, ))
    """May be populated with the names of files which contain data associated
    with each vertex in the mesh, that can be used to colour the mesh. When
    some vertex data has been succsessfully loaded, it can be accessed via
    the :meth:`getVertexData` method.

    This property is not currently populated by the ``MeshOpts`` class, but
    is used by sub-classes (e.g. :class:`.GiftiOpts`).
    """


    vertexDataIndex = props.Int(minval=0, maxval=0, default=0, clamped=True)
    """If :attr:`vertexData` is loaded, and has multiple data points per
    vertex (e.g. time series), this property controls the index into the
    data.
    """

    
    refImage = props.Choice()
    """A reference :class:`.Image` instance which the mesh coordinates are
    in terms of.

    For example, if this :class:`.TriangleMesh` represents the segmentation of
    a sub-cortical region from a T1 image, you would set the ``refImage`` to
    that T1 image.

    Any :class:`.Image` instance in the :class:`.OverlayList` may be chosen
    as the reference image.
    """

    
    useLut = props.Boolean(default=False)
    """If ``True``, and if some :attr:`vertexData` is loaded, the :attr:`lut`
    is used to colour vertex values instead of the :attr:`cmap` and
    :attr:`negativeCmap`.
    """


    lut = props.Choice()
    """If :attr:`useLut` is ``True``, a :class:`.LookupTable` is used to
    colour vertex data instead of the :attr:`cmap`/:attr:`negativeCmap`.
    """


    # This property is implicitly tightly-coupled to
    # the NiftiOpts.getTransform method - the choices
    # defined in this property are assumed to be valid
    # inputs to that method.
    coordSpace = props.Choice(('affine', 'pixdim', 'pixdim-flip', 'id'),
                              default='pixdim-flip')
    """If :attr:`refImage` is not ``None``, this property defines the
    reference image coordinate space in which the mesh coordinates are
    defined (i.e. voxels, scaled voxels, or world coordinates).

    =============== =========================================================
    ``affine``      The mesh coordinates are defined in the reference image
                    world coordinate system.
    
    ``id``          The mesh coordinates are defined in the reference image
                    voxel coordinate system.
    
    ``pixdim``      The mesh coordinates are defined in the reference image
                    voxel coordinate system, scaled by the voxel pixdims.
    
    ``pixdim-flip`` The mesh coordinates are defined in the reference image
                    voxel coordinate system, scaled by the voxel pixdims. If
                    the reference image transformation matrix has a positive
                    determinant, the X axis is flipped.
    =============== =========================================================

    The default value is ``pixdim-flip``, as this is the coordinate system
    used in the VTK sub-cortical segmentation model files output by FIRST.
    See also the :ref:`note on coordinate systems
    <volumeopts-coordinate-systems>`, and the :meth:`.NiftiOpts.getTransform`
    method.
    """


    def __init__(self, overlay, *args, **kwargs):
        """Create a ``MeshOpts`` instance. All arguments are passed through
        to the :class:`.DisplayOpts` constructor.
        """

        # Set a default colour 
        colour      = genMeshColour(overlay)
        self.colour = np.concatenate((colour, [1.0]))

        # ColourMapOpts.linkLowRanges defaults to
        # True, which is annoying for surfaces.
        self.linkLowRanges = False

        # A copy of the refImage property
        # value is kept here so, when it
        # changes, we can de-register from
        # the previous one.
        self.__oldRefImage = None

        # When the vertexData property is
        # changed, the data (and its min/max)
        # is loaded and stored in these 
        # attributes. See the __vertexDataChanged
        # method.
        self.__vertexData      = None
        self.__vertexDataRange = None

        nounbind = kwargs.get('nounbind', [])
        nounbind.extend(['refImage', 'coordSpace', 'transform'])
        kwargs['nounbind'] = nounbind
 
        fsldisplay.DisplayOpts  .__init__(self, overlay, *args, **kwargs)
        cmapopts  .ColourMapOpts.__init__(self)

        # TODO Only child instances should
        #      listen. Alternately, vertexData
        #      should be in nounbind.
        self.addListener('vertexData',
                         self.name,
                         self.__vertexDataChanged,
                         immediate=True) 

        # A number of callback functions are used to
        # keep the refImage, coordSpace and transform
        # properties consistent. Only the master
        # MeshOpts instance needs to register these
        # listeners.
        self.__registered = self.getParent() is None
        if self.__registered:
             
            self.overlayList.addListener('overlays',
                                         self.name,
                                         self.__overlayListChanged,
                                         immediate=True)

            self.addListener('refImage',
                             self.name,
                             self.__refImageChanged,
                             immediate=True)
            self.addListener('coordSpace',
                             self.name,
                             self.__coordSpaceChanged,
                             immediate=True)

            # The master MeshOpts instance also
            # keeps colour[3] and Display.alpha
            # consistent w.r.t. each other (see
            # also MaskOpts)
            self.display.addListener('alpha',
                                     self.name,
                                     self.__alphaChanged,
                                     immediate=True)
            self        .addListener('colour',
                                     self.name,
                                     self.__colourChanged,
                                     immediate=True) 
        
            self.__overlayListChanged()
            self.__updateBounds()

        # If a reference image has not
        # been set on the parent MeshOpts
        # instance, see  if there is a
        # suitable one in the overlay list. 
        if self.refImage is None:
            self.refImage = fsloverlay.findMeshReferenceImage(
                self.overlayList, self.overlay)


    def destroy(self):
        """Removes some property listeners, and calls the
        :meth:`.DisplayOpts.destroy` method.
        """

        if self.__registered:

            self.overlayList.removeListener('overlays', self.name)

            for overlay in self.overlayList:
                display = self.displayCtx.getDisplay(overlay)
                display.removeListener('name', self.name)

            if self.refImage is not None and \
               self.refImage in self.overlayList:
                opts = self.displayCtx.getOpts(self.refImage)
                opts.removeListener('transform',   self.name)
                opts.removeListener('customXform', self.name)

            self.display.removeListener('alpha',  self.name)
            self        .removeListener('colour', self.name)

        self.__oldRefImage = None
        self.__vertexData  = None
        
        cmapopts  .ColourMapOpts.destroy(self)
        fsldisplay.DisplayOpts  .destroy(self)


    def getDataRange(self):
        """Overrides the :meth:`.ColourMapOpts.getDisplayRange` method.
        Returns the display range of the currently selected
        :attr:`vertexData`, or ``(0, 1)`` if none is selected.
        """
        if self.__vertexDataRange is None: return (0, 1)
        else:                              return self.__vertexDataRange
            

    def getVertexData(self):
        """Returns the :attr:`.MeshOpts.vertexData`, if some is loaded.
        Returns ``None`` otherwise.
        """
        return self.__vertexData


    def vertexDataLen(self):
        """Returns the length (number of data points per vertex) of the
        currently selected :attr:`vertexData`, or ``0`` if no vertex data is
        selected.
        """
        
        if self.__vertexData is None:
            return 0
        
        elif len(self.__vertexData.shape) == 1:
            return 1
        
        else:
            return self.__vertexData.shape[1]


    def addVertexDataOptions(self, paths):
        """Adds the given sequence of paths as options to the
        :attr:`vertexData` property. It is assumed that the paths refer
        to valid vertex data files for the overlay associated with this
        ``MeshOpts`` instance.
        """

        vdataProp = self.getProp('vertexData')
        newPaths  = paths
        paths     = vdataProp.getChoices(instance=self)
        paths     = paths + [p for p in newPaths if p not in paths]
        
        vdataProp.setChoices(paths, instance=self)

    
    def getConstantColour(self):
        """Returns the current :attr::`colour`, adjusted according to the
        current :attr:`.Display.brightness`, :attr:`.Display.contrast`, and
        :attr:`.Display.alpha`.
        """

        display = self.display

        colour = list(fslcmaps.applyBricon(
            self.colour[:3],
            display.brightness / 100.0,
            display.contrast   / 100.0))
        
        colour.append(display.alpha / 100.0)

        return colour


    def getReferenceImage(self):
        """Overrides :meth:`.DisplayOpts.getReferenceIamge`.

        If a :attr:`refImage` is selected, it is returned. Otherwise,``None``
        is returned.
        """
        return self.refImage

    
    def getCoordSpaceTransform(self):
        """Returns a transformation matrix which can be used to transform
        the :class:`.TriangleMesh` vertex coordinates into the display
        coordinate system.

        If no :attr:`refImage` is selected, this method returns an identity
        transformation.
        """

        if self.refImage is None:
            return np.eye(4)

        opts = self.displayCtx.getOpts(self.refImage)

        return opts.getTransform(self.coordSpace, opts.transform)
    

    def displayToStandardCoordinates(self, coords):
        """Transforms the given coordinates into a standardised coordinate
        system specific to the overlay associated with this ``MeshOpts``
        instance.

        The coordinate system used is the coordinate system in which the
        :class:`.TriangleMesh` vertices are defined.
        """
        if self.refImage is None:
            return coords

        opts = self.displayCtx.getOpts(self.refImage)
        
        return opts.transformCoords(coords, opts.transform, self.coordSpace)

    
    def standardToDisplayCoordinates(self, coords):
        """Transforms the given coordinates from standardised coordinates
        into the display coordinate system - see
        :meth:`displayToStandardCoordinates`.
        """
        if self.refImage is None:
            return coords

        opts = self.displayCtx.getOpts(self.refImage)
        
        return opts.transformCoords(coords, self.coordSpace, opts.transform)

    
    def __cacheCoords(self,
                      refImage=-1,
                      coordSpace=None,
                      transform=None,
                      customXform=None):
        """Caches the current :attr:`.DisplayContext.location` in standardised
        coordinates see the :meth:`.DisplayContext.cacheStandardCoordinates`
        method).

        This method is called whenever the :attr:`refImage` or
        :attr:`coordSpace` properties change and, if a ``refImage`` is
        specified, whenever the :attr:`.NiftiOpts.transform` or
        :attr:`.NiftiOpts.customXform` properties change.

        :arg refImage:    Reference image to use to calculate the coordinates.
                          If ``-1`` the :attr:`refImage` is used (``-1`` is
                          used as the default value instead of ``None`` because
                          the latter is a valid value for ``refImage``).
        
        :arg coordSpace:  Coordinate space value to use - if ``None``, the
                          :attr:`coordSpace` is used.
        
        :arg transform:   Transform to use - if ``None``, and a ``refImage`` is
                          defined, the :attr:`.NiftiOpts.transform` value is
                          used.
        
        :arg customXform: Custom transform to use (if
                          ``transform=custom``). If ``None``, and a
                          ``refImage`` is defined, the
                          :attr:`.NiftiOpts.customXform` value is used.
        """

        if refImage   is -1:   refImage   = self.refImage
        if coordSpace is None: coordSpace = self.coordSpace

        if refImage is None:
            coords = self.displayCtx.location.xyz
            
        else:
            refOpts = self.displayCtx.getOpts(refImage)

            if transform   is None: transform   = refOpts.transform
            if customXform is None: customXform = refOpts.customXform

            # TODO if transform == custom, we 
            # have to use the old custom xform
            
            coords  = refOpts.transformCoords(self.displayCtx.location.xyz,
                                              transform,
                                              coordSpace)
        
        self.displayCtx.cacheStandardCoordinates(self.overlay, coords)


    def __transformChanged(self, value, valid, ctx, name):
        """Called when the :attr:`.NiftiOpts.transfrom` or
        :attr:`.NiftiOpts.customXform` properties of the current
        :attr:`refImage` change. Calls :meth:`__updateBounds`.
        """

        refOpts = ctx

        if   name == 'transform':
            transform   = refOpts.getLastValue('transform')
            customXform = refOpts.customXform
        elif name == 'customXform': 
            transform   = refOpts.transform
            customXform = refOpts.getLastValue('customXform')

        self.__cacheCoords(transform=transform, customXform=customXform)
        self.__updateBounds()


    def __coordSpaceChanged(self, *a):
        """Called when the :attr:`coordSpace` property changes.
        Calls :meth:`__updateBounds`.
        """

        oldValue = self.getLastValue('coordSpace')

        if oldValue is None:
            oldValue = self.coordSpace

        self.__cacheCoords(coordSpace=oldValue)
        self.__updateBounds()


    def __refImageChanged(self, *a):
        """Called when the :attr:`refImage` property changes.

        If a new reference image has been specified, removes listeners from
        the old one (if necessary), and adds listeners to the
        :attr:`.NiftiOpts.transform` and :attr:`.NiftiOpts.customXform`
        properties associated with the new image. Calls
        :meth:`__updateBounds`.
        """
        
        oldValue = self.getLastValue('refImage')

        # The reference image may have been
        # removed from the overlay list
        if oldValue not in self.overlayList:
            oldValue = None

        self.__cacheCoords(refImage=oldValue) 

        # TODO You are not tracking changes to the
        # refImage overlay type -  if this changes,
        # you will need to re-bind to the transform
        # property of the new DisplayOpts instance

        if self.__oldRefImage is not None and \
           self.__oldRefImage in self.overlayList:
            
            opts = self.displayCtx.getOpts(self.__oldRefImage)
            opts.removeListener('transform',   self.name)
            opts.removeListener('customXform', self.name)

        self.__oldRefImage = self.refImage

        if self.refImage is not None:
            opts = self.displayCtx.getOpts(self.refImage)
            opts.addListener('transform',
                             self.name,
                             self.__transformChanged,
                             immediate=True)
            opts.addListener('customXform',
                             self.name,
                             self.__transformChanged,
                             immediate=True)

        self.__updateBounds()


    def __updateBounds(self):
        """Called whenever any of the :attr:`refImage`, :attr:`coordSpace`, 
        or :attr:`transform` properties change.

        Updates the :attr:`.DisplayOpts.bounds` property accordingly.
        """

        lo, hi = self.overlay.getBounds()
        xform  = self.getCoordSpaceTransform()

        lohi   = transform.transform([lo, hi], xform)
        lohi.sort(axis=0)
        lo, hi = lohi[0, :], lohi[1, :]

        oldBounds = self.bounds
        self.bounds = [lo[0], hi[0], lo[1], hi[1], lo[2], hi[2]]

        # Horrible hack here.

        # The coordSpace/refImage/transform property
        # change may not result in a change to the
        # bound values. But listeners of the bounds
        # property need to be notified regardless, as
        # the model space has changed (e.g. it may
        # have just been flipped along an axis).
        # For example, the OrthoPanel needs to refresh
        # its orientation labels.
        # 
        # This method is only called on the 'master'
        # MeshOpts instance - the bounds on child
        # instances are synced automatically. So we
        # have to force notification of all bounds
        # listeners on the child instances.
        #
        # Hopefully in the future I will come up with
        # a solution to these horrible parent-child
        # discrepancies.
        if oldBounds == self.bounds:
            children = self.getChildren()
            for c in [self] + children:
                c.propNotify('bounds')
            
    
    def __overlayListChanged(self, *a):
        """Called when the overlay list changes. Updates the :attr:`refImage`
        property so that it contains a list of overlays which can be
        associated with the mesh.
        """
        
        imgProp  = self.getProp('refImage')
        imgVal   = self.refImage
        overlays = self.displayCtx.getOrderedOverlays()

        # the overlay for this MeshOpts
        # instance has been removed
        if self.overlay not in overlays:
            self.overlayList.removeListener('overlays', self.name)
            return

        imgOptions = [None]

        for overlay in overlays:

            # The overlay must be a Nifti instance.
            if not isinstance(overlay, fslimage.Nifti):
                continue

            imgOptions.append(overlay)

            display = self.displayCtx.getDisplay(overlay)
            display.addListener('name',
                                self.name,
                                self.__overlayListChanged,
                                overwrite=True)

        # The previous refImage may have
        # been removed from the overlay list
        if imgVal in imgOptions: self.refImage = imgVal
        else:                    self.refImage = None
 
        imgProp.setChoices(imgOptions, instance=self)


    def __vertexDataChanged(self, *a):
        """Called when the :attr:`vertexData` property changes. Attempts to
        load the data if possible. The data may subsequently be retrieved
        via the :meth:`getVertexData` method.
        """

        vdata      = None
        vdataRange = None

        try:
            if self.vertexData is not None:
                vdata      = self.overlay.getVertexData(self.vertexData)
                vdataRange = np.nanmin(vdata), np.nanmax(vdata)

                if len(vdata.shape) == 1:
                    vdata = vdata.reshape(-1, 1)
                
        except Exception as e:

            # TODO show a warning
            log.warning('Unable to load vertex data from {}: {}'.format(
                self.vertexData, e, exc_info=True))

            vdata      = None
            vdataRange = None

        self.__vertexData      = vdata
        self.__vertexDataRange = vdataRange
        
        if vdata is not None: npoints = vdata.shape[1]
        else:                 npoints = 1    

        self.vertexDataIndex = 0 
        self.setConstraint('vertexDataIndex', 'maxval', npoints - 1)

        self.updateDataRange()


    def __colourChanged(self, *a):
        """Called when :attr:`.colour` changes. Updates :attr:`.Display.alpha`
        from the alpha component.
        """ 

        alpha = self.colour[3] * 100

        log.debug('Propagating MeshOpts.colour[3] to '
                  'Display.alpha [{}]'.format(alpha))

        with props.skip(self.display, 'alpha', self.name):
            self.display.alpha = alpha


    def __alphaChanged(self, *a):
        """Called when :attr:`.Display.alpha` changes. Updates the alpha
        component of :attr:`.colour`.
        """

        alpha      = self.display.alpha / 100.0
        r, g, b, _ = self.colour

        log.debug('Propagating Display.alpha to MeshOpts.'
                  'colour[3] [{}]'.format(alpha))

        with props.skip(self, 'colour', self.name):
            self.colour = r, g, b, alpha
